#ifndef _LARGEINT_H_
#define _LARGEINT_H_

#include <stdint.h>
#include <memory.h>
#include <string.h>

struct LI_Unit
{
	uint16_t data;	// data chunk
};

template <int T>
class LargeInt
{
private:
	uint16_t	m_BitMask[16];

	// hex char to int
	uint8_t		htoi(char val)
	{
		switch(val)
		{
		case '1':
			return 1;
		case '2':
			return 2;
		case '3':
			return 3;
		case '4':
			return 4;
		case '5':
			return 5;
		case '6':
			return 6;
		case '7':
			return 7;
		case '8':
			return 8;
		case '9':
			return 9;
		case 'a':
		case 'A':
			return 10;
		case 'b':
		case 'B':
			return 11;
		case 'c':
		case 'C':
			return 12;
		case 'd':
		case 'D':
			return 13;
		case 'e':
		case 'E':
			return 14;
		case 'f':
		case 'F':
			return 15;
		default:
			return 0;
		}
	}

public:
	// the large int value
	LI_Unit		value[T];

	LargeInt()
	{
		null();

		// prepare bitmasks
		for (int i = 0; i < 16; ++i)
			m_BitMask[i] = 1 << i;
	};
	LargeInt(uint16_t inVal)
	{
		null();

		// prepare bitmasks
		for (int i = 0; i < 16; ++i)
			m_BitMask[i] = 1 << i;

		value[0].data = inVal;
	};
	~LargeInt() { };

	LargeInt<T> operator+ (LargeInt<T>& inVal)
	{
		// result
		LargeInt<T> result;

		// sequential addition
		uint32_t tmp;
		uint8_t rem = 0;
		uint16_t mask = ~((uint16_t)0);
		for (int i = 0; i < T; ++i)
		{
			// the addition
			tmp = value[i].data + inVal.value[i].data + rem;

			// store the result
			result.value[i].data = tmp & mask;

			// store result
			rem = tmp >> 16;
		}

		return result;
	};

	LargeInt<T> operator+ (uint16_t inOther)
	{
		return *this + LargeInt<T>(inOther);
	};

	LargeInt<T> operator- (LargeInt<T>& inVal)
	{
		// result
		LargeInt<T> result;

		// sequential subtraction
		uint16_t rem = 0;
		uint16_t tmp;
		for (int i = 0; i < T; ++i)
		{
			tmp = inVal.value[i].data + rem;

			result.value[i].data = value[i].data - tmp;

			rem = value[i].data < tmp;
		}

		if (rem)
			result.null();

		return result;
	};

	LargeInt<T> operator- (uint16_t inOther)
	{
		return *this - LargeInt<T>(inOther);
	};

	void operator -= (LargeInt<T>& inOther)
	{
		*this = *this - inOther;
	};

	void operator -= (uint16_t inOther)
	{
		*this = *this - LargeInt<T>(inOther);
	};

	LargeInt<T> operator * (LargeInt<T>& inOther)
	{
		// result
		LargeInt<T> result;

		// sequential multiplication
		for (int i = 0; i < T; ++i)
			for (int bit = 0; bit < 16; ++bit)
				if (inOther.value[i].data & this->m_BitMask[bit])
					result += *this << (bit + 16 * i);

		return result;
	};

	void operator *= (LargeInt<T>& inOther)
	{
		*this = *this * inOther;
	};

	LargeInt<T> operator * (uint16_t inOther)
	{
		// result
		LargeInt<T> result;

		// sequential multiplication
		for (int bit = 0; bit < 16; ++bit)
			if (inOther & m_BitMask[bit])
				result += *this << bit;

		return result;
	};

	void operator *= (uint16_t inOther)
	{
		*this = *this * inOther;
	};

	LargeInt<T> operator << (uint16_t val)
	{
		// result
		LargeInt<T> result;

		// first check how many LI_Units are going to be empty
		uint16_t num = val / 16;
		if (num >= T)
			return result;

		// start at the highest bits and work down
		uint16_t rem = val % 16;
		for (int i = T-1; i > num; --i)
		{
			result.value[i].data = (value[i-num].data << rem) | (value[i-1-num].data >> (16 - rem));
		}
		result.value[num].data = value[0].data << rem;
		for (int i = 0; i < num; ++i)
			result.value[i].data = 0;

		return result;
	};

	void operator <<= (uint16_t val)
	{
		*this = *this << val;
	}

	LargeInt<T> operator >> (uint16_t val)
	{
		LargeInt<T> result;

		// first check how many LI_Units are going to be empty
		uint16_t num = val / 16;
		if (num > T)
			return result;

		// start at the lowest bits and work up
		uint16_t rem = val % 16;
		for (int i = 0; i < (T-num)-1; ++i)
		{
			result.value[i].data = (this->value[i+num].data >> rem) | (this->value[i+num+1].data << (16 - rem));
		}
		result.value[T-num-1].data = this->value[T-1].data >> rem;
		for (int i = (T-num); i < T; ++i)
			result.value[i].data = 0;

		return result;
	};

	void operator >>= (uint16_t val)
	{
		*this = *this >> val;
	}

	LargeInt<T> operator = (const char* inVal)
	{
		// check prefix
		if (strncmp(inVal, "0x", 2))
			return *this;

		null();

		int len = strlen(inVal) - 1;
		for (int i = 0; i < T && len > 0; ++i)
		{
			for (int x = 0; x < 4; ++x)
			{
				value[i].data += htoi( inVal[len - x] ) << (x * 4);
			}
			len -= 4;
		}

		return *this;
	};

	LargeInt<T> operator= (LargeInt<T>& inOther)
	{
		memcpy(value, inOther.value, sizeof(LI_Unit) * T);
		return *this;
	};

	void operator += (LargeInt<T>& inOther)
	{
		*this = *this + inOther;
	};

	void operator += (uint16_t inOther)
	{
		*this = *this + LargeInt<T>(inOther);
	};

	bool operator== (LargeInt<T>& inOther)
	{
		for (int i = 0; i < T; ++i)
			if (value[i].data != inOther.value[i].data)
				return false;

		return true;
	};

	bool operator== (uint16_t inOther)
	{
		for (int i = T-1; i > 0; --i)
			if (value[i].data)
				return false;

		return value[0].data == inOther;
	};

	LargeInt<T> operator& (LargeInt<T>& inOther)
	{
		LargeInt<T> result;

		for (int i = 0; i < T; ++i)
			result.value[i].data = this->value[i].data & inOther.value[i].data;

		return result;
	};

	LargeInt<T> operator& (uint16_t inOther)
	{
		LargeInt<T> result;

		result.value[0].data = value[0].data & inOther;

		return result;
	};

	void operator &= (LargeInt<T>& inOther)
	{
		for (int i = 0; i < T; ++i)
			this->value[i].data = this->value[i].data & inOther.value[i].data;
	};

	void operator &= (uint16_t inOther)
	{
		// and operation
		uint16_t tmp = value[0].data & inOther;

		// reset this int
		null();

		// set result of and operation
		value[0].data = tmp;
	};

	LargeInt<T> operator| (LargeInt<T>& inOther)
	{
		LargeInt<T> result;

		for (int i = 0; i < T; ++i)
			result.value[i].data = this->value[i].data | inOther.value[i].data;

		return result;
	};

	LargeInt<T> operator| (uint16_t inOther)
	{
		// create a copy
		LargeInt<T> result;
		result = *this;

		// or operation
		result.value[0].data |= inOther;

		return result;
	};

	void operator |= (LargeInt<T>& inOther)
	{
		for (int i = 0; i < T; ++i)
			this->value[i].data = this->value[i].data | inOther.value[i].data;
	};

	void operator |= (uint16_t inOther)
	{
		value[0].data |= inOther;
	};

	bool operator > (LargeInt<T>& inOther)
	{
		for (int i = T-1; i >= 0; --i)
		{
			if (this->value[i].data < inOther.value[i].data)
				return false;
			else if (this->value[i].data > inOther.value[i].data)
				return true;
		}

		return false;
	};

	bool operator > (uint16_t inOther)
	{
		for (int i = T-1; i > 0; --i)
			if (value[i].data)
				return true;

		return value[0].data > inOther;
	}

	bool operator >= (LargeInt<T>& inOther)
	{
		for (int i = T-1; i >= 0; --i)
		{
			if (this->value[i].data < inOther.value[i].data)
				return false;
			else if (this->value[i].data > inOther.value[i].data)
				return true;
		}

		return true;
	};

	bool operator >= (uint16_t inOther)
	{
		for (int i = T-1; i > 0; --i)
			if (value[i].data)
				return true;

		return value[0].data >= inOther;
	}

	bool operator < (LargeInt<T>& inOther)
	{
		for (int i = T-1; i >= 0; --i)
		{
			if (this->value[i].data > inOther.value[i].data)
				return false;
			else if (this->value[i].data < inOther.value[i].data)
				return true;
		}

		return false;
	};

	bool operator < (uint16_t inOther)
	{
		for (int i = T-1; i > 0; --i)
			if (value[i].data)
				return false;

		return value[0].data < inOther;
	}

	bool operator <= (LargeInt<T>& inOther)
	{
		for (int i = T-1; i >= 0; --i)
		{
			if (this->value[i].data > inOther.value[i].data)
				return false;
			else if (this->value[i].data < inOther.value[i].data)
				return true;
		}

		return true;
	};

	bool operator <= (uint16_t inOther)
	{
		for (int i = T-1; i > 0; --i)
			if (value[i].data)
				return false;

		return value[0].data <= inOther;
	}

	LargeInt<T> operator / (LargeInt<T>& inOther)
	{
		LargeInt<T> result;

		if (*this >= inOther)
		{
			// working copy
			LargeInt<T> me = *this,
						other = inOther;	

			// get the highest bits to determine where to start
			uint16_t	highOwn		= me.highestBit(),
						highOther	= other.highestBit();

			// shift the divisor by the difference to the left
			other <<= (highOwn - highOther);

			// sequential division
			for (int i = 0; i <= highOwn - highOther; ++i)
			{
				// shift the result
				result <<= 1;

				if (me >= other)
				{
					// 1
					result |= 1;

					// substract the other
					me -= other;
				}

				// shift divisor to the right
				other >>= 1;
			}
		}

		return result;
	};

	LargeInt<T> operator / (uint16_t inOther)
	{
		return *this / LargeInt<T>(inOther);
	};

	void operator /= (LargeInt<T>& inOther)
	{
		*this = *this / inOther;
	};

	void operator /= (uint16_t inOther)
	{
		*this = *this / LargeInt<T>(inOther);
	};

	LargeInt<T> operator % (LargeInt<T>& inOther)
	{
		LargeInt<T> result;

		if (*this >= inOther)
		{
			// working copy
			LargeInt<T> me = *this,
						other = inOther;	

			// get the highest bits to determine where to start
			uint16_t	highOwn		= me.highestBit(),
						highOther	= other.highestBit();

			// shift the divisor by the difference to the left
			other <<= (highOwn - highOther);

			// sequential division
			for (int i = 0; i <= highOwn - highOther; ++i)
			{
				if (me >= other)
				{
					// substract the other
					me -= other;
				}

				// shift divisor to the right
				other >>= 1;
			}

			// the remainder of the operation
			result = me;
		}
		else
			result = *this;

		return result;
	};

	LargeInt<T> operator % (uint16_t inOther)
	{
		return *this % LargeInt<T>(inOther);
	};

	void operator %= (LargeInt<T>& inOther)
	{
		*this = *this % inOther;
	};

	void operator %= (uint16_t inOther)
	{
		*this = *this % LargeInt<T>(inOther);
	};

	// prints the number in hex format
	void print()
	{
		printf("0x ");
		for (int i = T-1; i >= 0; --i)
		{
			for (int x = 3; x >= 0; --x)
				printf("%x", (value[i].data >> (x * 4)) & 15);
			printf(" ");
		}
	};

	void null()
	{
		// zero buffer
		memset(this->value, 0, sizeof(LI_Unit) * T);
	};

	uint16_t highestBit()
	{
		for (int i = T-1; i >= 0; --i)
			if (this->value[i].data)
				for (int bit = 15; bit >= 0; --bit)
					if (this->value[i].data & this->m_BitMask[bit])
						return i * 16 + bit;

		return 0;
	};
};

template<int T>
LargeInt<T>::LargeInt();



#endif